def bspline( p, *knots, **kw ):
    repeat = kw.get( 'repeat', 0 )
    loop = kw.get( 'loop', 0 )

    if loop:
        if len(knots) < 3:
            raise ValueError, "must have at least 3 knots for b-spline loop"
        knots = knots + knots[:3]
    else:
        if repeat:
            if len(knots) < 1:
                raise ValueError, "must have at least 1 knot for b-spline with repeat"
            knots = knots[:1] * 2 + knots + knots[-1:] * 2
        else:
            if len(knots) < 4:
                raise ValueError, "must have at least 4 knots for b-spline without repeat"

    for i in range(len(knots)-3):
        (x0,y0), (x1,y1), (x2,y2), (x3,y3) = knots[i:i+4]

        if i == 0:
            xb0 = (x0 + 4 * x1 + x2) / 6.0
            yb0 = (y0 + 4 * y1 + y2) / 6.0
            p.moveto( xb0, yb0 )

        xb1 = (2 * x1 + x2) / 3.0
        xb2 = (x1 + 2 * x2) / 3.0
        xb3 = (x1 + 4 * x2 + x3) / 6.0

        yb1 = (2 * y1 + y2) / 3.0
        yb2 = (y1 + 2 * y2) / 3.0
        yb3 = (y1 + 4 * y2 + y3) / 6.0

        p.curveto( xb1, yb1, xb2, yb2, xb3, yb3 )

    if loop:
        p.closepath()

    return p

def catmullrom( p, *knots, **kw ):
    tension = kw.get( 'tension', 0.5 )
    loop = kw.get( 'loop', 0 )

    if loop:
        if len(knots) < 2:
            raise ValueError, "must have at least 2 knots for catmull-rom loop"

        d = []
        knots = knots + knots[:2]
        for i in range(1,len(knots)-1):
            d.append( (tension * (knots[i+1][0] - knots[i-1][0]), tension * (knots[i+1][1] - knots[i-1][1])) )
        d.insert( 0, d[-1] )
        
    else:
        if len(knots) < 2:
            raise ValueError, "must have at least 2 knots for catmull-rom spline"

        d = [ (tension * (knots[1][0] - knots[0][0]), tension * (knots[1][1] - knots[0][1])) ]
        for i in range(1,len(knots)-1):
            d.append( (tension * (knots[i+1][0] - knots[i-1][0]), tension * (knots[i+1][1] - knots[i-1][1])) )
        l = len(knots)-1
        d.append( (tension * (knots[l][0] - knots[l-1][0]), tension * (knots[l][1] - knots[l-1][1])) )

    for i in range(len(knots)-1-(not not loop)):
        (x,y), (xn,yn) = knots[i:i+2]
        (xd,yd), (xnd,ynd) = d[i:i+2]
        
        if i == 0:
            p.moveto( x, y )
            
        p.curveto( x + xd / 3.0, y + yd / 3.0, xn - xnd / 3.0, yn - ynd / 3.0, xn, yn )

    if loop:
        p.closepath()

    return p

def c2interpolating( p, *knots, **kw ):
    loop = kw.get( 'loop', 0 )

    n = len(knots)
        
    if loop:
        if n < 3:
            raise ValueError, "must have at least 3 knots for c2-interpolating loop"
        
        deltax = []
        deltay = []
        for next, prev in zip(range(n)[1:]+[0], [n-1] + range(n)[:-1]):
            deltax.append( 3.0 * (knots[next][0] - knots[prev][0]) )
            deltay.append( 3.0 * (knots[next][1] - knots[prev][1]) )

        ##
        ## forward elimination
        ##

        gamma = [ 0.25 ]
        sigma = [ 0.25 ]
        deltax[0] /= 4.0
        deltay[0] /= 4.0

        for i in range(1,n-2):
            g = 4.0 - gamma[-1]

            deltax[i] = (deltax[i] - deltax[i-1]) / g
            deltay[i] = (deltay[i] - deltay[i-1]) / g

            gamma.append( 1.0 / g )
            sigma.append( -sigma[-1] / g )

        deltax[n-2] = (deltax[n-2] - deltax[n-3]) / (4.0 - gamma[-1])
        deltay[n-2] = (deltay[n-2] - deltay[n-3]) / (4.0 - gamma[-1])
        gamma.append( (1.0 - sigma[-1]) / (4.0 - gamma[-1]) )
        sigma.append( 0.0 )

        ##
        ## bottom row
        ##

        phi = 1.0
        tau = 4.0
        for i in range(n-2):
            deltax[n-1] = deltax[n-1] - phi * deltax[i]
            deltay[n-1] = deltay[n-1] - phi * deltay[i]

            tau = tau - phi * sigma[i]
            phi = -phi * gamma[i]

        deltax[n-1] = deltax[n-1] - (phi+1) * deltax[n-2]
        deltay[n-1] = deltay[n-1] - (phi+1) * deltay[n-2]

        tau = tau - gamma[n-2] * (phi+1)

        ##
        ## back substitution
        ##

        d = [ (deltax[n-1] / tau, deltay[n-1] / tau) ]
        d.append( (deltax[n-2] - gamma[n-2] * d[0][0],
                   deltay[n-2] - gamma[n-2] * d[0][1]) )

        for i in range(n-3,-1,-1):
            d.append( (deltax[i] - gamma[i] * d[-1][0] - sigma[i] * d[0][0],
                       deltay[i] - gamma[i] * d[-1][1] - sigma[i] * d[0][1]) )

        d.reverse()

        knots = knots + ( knots[0], )
        d.append( d[0] )
        
    else:
        if n < 3:
            raise ValueError, "must have at least 3 knots for c2-interpolating spline"
        
        ##
        ## forward elimination
        ##
        
        gamma = [ 0.5 ]
        deltax = [ 3 * gamma[-1] * (knots[1][0] - knots[0][0]) ]
        deltay = [ 3 * gamma[-1] * (knots[1][1] - knots[0][1]) ]

        for i in range(1,len(knots)-1):
            g = 1.0 / ( 4.0 - gamma[-1] )
            gamma.append( g )

            xp, yp = knots[i-1]
            xn, yn = knots[i+1]
            
            deltax.append( g * (3 * (xn-xp) - deltax[-1]) )
            deltay.append( g * (3 * (yn-yp) - deltay[-1]) )
            
        g = 1.0 / ( 2.0 - gamma[-1] )
        gamma.append( g )

        xp, yp = knots[i]
        xn, yn = knots[i+1]

        deltax.append( g * (3 * (xn-xp) - deltax[-1]) )
        deltay.append( g * (3 * (yn-yp) - deltay[-1]) )

        ##
        ## back substitution
        ##

        d = [ (deltax[-1], deltay[-1]) ]
        for i in range(len(knots)-2, -1, -1):
            dx, dy = d[-1]
            d.append( (deltax[i] - gamma[i] * dx, deltay[i] - gamma[i] * dy) )
        d.reverse()

    for i in range(n-(not loop)):
        (x,y), (xn,yn) = knots[i:i+2]
        (xd,yd), (xnd,ynd) = d[i:i+2]
        
        if i == 0:
            p.moveto( x, y )
            
        p.curveto( x + xd / 3.0, y + yd / 3.0, xn - xnd / 3.0, yn - ynd / 3.0, xn, yn )

    if loop:
        p.closepath()
        
    return p

        
        
if __name__ == '__main__':
    from pres.combined.diagrams import *

    def test( x = (SCALAR, 0, 1),
              tension = (SCALAR, 0, 1, 0.5),
              loop = (BOOLEAN),
              repeat = (BOOLEAN),
              curvetype = (INTEGER, 0, 2),
              ):
        pts = ((0,0), (0,0.5+x), (1,1-2*x), (1+x,0), (2,2-x))

        curvetype = (bspline, catmullrom, c2interpolating)[curvetype]

        p = Path()
        curvetype( p, *pts, **{'loop' : loop, 'tension' : tension, 'repeat' : repeat } )

        color( blue )
        widestroke( p, 0.05 )

        color( red )
        for p in pts:
            dot( 0.04, p )
    test.cameras = {
       'default' : (-0.58533739554457953, -0.61816759665814358, 3.0525184179211178, 0, 1),
    }

    test_diagrams( test, clear_color = white )

        
    
